@@ -14,19 +14,19 @@
 #  limitations under the License.
 
 import base64
-from typing import Optional, Union, List, Dict
+import dataclasses
+import json
+import traceback
+from typing import Any, Dict, List, Optional, Union
+
 import requests
 import urllib3
-from . import questions
-from .util import (
-    XmlExportConfig,
-    ImportConfig,
-    close_program,
-    spin_spinner,
-    AbstractAction,
-)
 from questionary import print as pprint
-import json
+
+from . import questions
+from .config_model import CliReporterConfig, Configuration, ExecutionResultsImportOptions
+from .log import logger
+from .util import AbstractAction, ImportConfig, XmlExportConfig, close_program, spin_spinner
 
 
 class Connection:
@@ -66,9 +66,7 @@ class Connection:
         self._session.headers.update(
             {"Content-Type": "application/vnd.testbench+json; charset=utf-8"}
         )
-        self._session.hooks = {
-            "response": lambda r, *args, **kwargs: r.raise_for_status()
-        }
+        self._session.hooks = {"response": lambda r, *args, **kwargs: r.raise_for_status()}
         self._session.mount("http://", TimeoutHTTPAdapter(self.connection_timeout))
         self._session.mount("https://", TimeoutHTTPAdapter(self.connection_timeout))
         self._session.verify = self.verify_ssl
@@ -77,20 +75,14 @@ class Connection:
     def close(self):
         self.session.close()
 
-    def export(self) -> dict:
-        basic_auth = base64.b64encode(
-            f"{self.loginname}:{self.password}".encode("utf-8")
-        ).decode()
-        return {
-            "server_url": self.server_url,
-            "verify": self.session.verify,
-            "basicAuth": basic_auth,
-            "actions": [
-                action_export
-                for action_export in (action.export() for action in self.action_log)
-                if action_export is not None
-            ],
-        }
+    def export(self) -> Configuration:
+        basic_auth = base64.b64encode(f"{self.loginname}:{self.password}".encode("utf-8")).decode()
+        return Configuration(
+            server_url=self.server_url,
+            verify=self.session.verify,
+            basicAuth=basic_auth,
+            actions=[action.export() for action in self.action_log if action.export() is not None],
+        )
 
     def add_action(self, action: AbstractAction):
         self.action_log.append(action)
@@ -107,7 +99,7 @@ class Connection:
 
     def check_is_working(self) -> bool:
         response = self.session.get(
-            self.server_url + "projects",
+            f"{self.server_url}projects",
             params={
                 "includeTOVs": "false",
                 "includeCycles": "false",
@@ -120,7 +112,7 @@ class Connection:
 
     def get_all_projects(self) -> Dict:
         all_projects = self.session.get(
-            self.server_url + "projects",
+            f"{self.server_url}projects",
             params={"includeTOVs": "true", "includeCycles": "true"},
         ).json()
         all_projects["projects"].sort(key=lambda proj: proj["name"].casefold())
@@ -128,7 +120,7 @@ class Connection:
 
     def get_all_filters(self) -> List[dict]:
         all_filters = self.session.get(
-            self.server_url + "filters",
+            f"{self.server_url}filters",
         )
 
         return all_filters.json()
@@ -138,9 +130,7 @@ class Connection:
     ) -> bytes:
         if filters is None:
             filters = []
-        job_id = self.trigger_xml_report_generation(
-            tov_key, cycle_key, reportRootUID, filters
-        )
+        job_id = self.trigger_xml_report_generation(tov_key, cycle_key, reportRootUID, filters)
         report_tmp_name = self.wait_for_tmp_xml_report_name(job_id)
         report = self.get_xml_report_data(report_tmp_name)
 
@@ -160,21 +150,18 @@ class Connection:
             filters = []
 
         if reportRootUID and reportRootUID != "ROOT":
-            report_config["reportRootUID"] = reportRootUID
-        report_config["filters"]: filters
-
+            report_config.reportRootUID = reportRootUID
+        report_config.filters = filters
         if cycle_key and cycle_key != "0":
             response = self.session.post(
-                self.server_url + "cycle/" + cycle_key + "/xmlReport",
-                json=report_config,
+                f"{self.server_url}cycle/{cycle_key}/xmlReport",
+                json=dataclasses.asdict(report_config),
             )
         else:
             response = self.session.post(
-                self.server_url + "tovs/" + tov_key + "/xmlReport",
-                json=report_config,
+                f"{self.server_url}tovs/{tov_key}/xmlReport",
+                json=dataclasses.asdict(report_config),
             )
-        if response.status_code != requests.codes.accepted:
-            raise AssertionError(f"{response.status_code} {response.text}")
         return response.json()["jobID"]
 
     def wait_for_tmp_xml_report_name(self, job_id: str) -> str:
@@ -202,7 +189,7 @@ class Connection:
 
     def get_xml_report_data(self, report_tmp_name: str) -> bytes:
         report = self.session.get(
-            self.server_url + "xmlReport/" + report_tmp_name,
+            f"{self.server_url}xmlReport/{report_tmp_name}",
         )
 
         return report.content
@@ -216,7 +203,7 @@ class Connection:
 
     def get_all_members_of_project(self, project_key: str) -> List[dict]:
         all_project_members = self.session.get(
-            self.server_url + "project/" + project_key + "/members",
+            f"{self.server_url}project/{project_key}/members",
         )
 
         return all_project_members.json()
@@ -224,7 +211,7 @@ class Connection:
     def upload_execution_results(self, results_file_base64: str) -> str:
         try:
             serverside_file_name = self.session.post(
-                self.server_url + "executionResultsUpload",
+                f"{self.server_url}executionResultsUpload",
                 json={
                     "data": results_file_base64,
                 },
@@ -240,34 +227,32 @@ class Connection:
         serverside_file_name: str,
         default_tester: str,
         filters: List[Dict[str, str]],
-        import_config: Dict = ImportConfig["Typical"],
+        import_config: Optional[ExecutionResultsImportOptions] = None,
     ) -> str:
-
-        import_config["fileName"] = serverside_file_name
-        import_config["filters"] = filters
+        if import_config is None:
+            import_config = ImportConfig["Typical"]
+        import_config.fileName = serverside_file_name
+        import_config.filters = filters
         if default_tester:
-            import_config["defaultTester"] = default_tester
+            import_config.defaultTester = default_tester
         if report_root_uid and report_root_uid != "ROOT":
-            import_config["reportRootUID"] = report_root_uid
+            import_config.reportRootUID = report_root_uid
 
         try:
             job_id = self.session.post(
-                self.server_url + "cycle/" + cycle_key + "/executionResultsImport",
-                headers={
-                    "Accept": "application/zip",
-                },
-                json=import_config,
+                f"{self.server_url}cycle/{cycle_key}/executionResultsImport",
+                headers={"Accept": "application/zip"},
+                json=dataclasses.asdict(import_config),
             )
             return job_id.json()["jobID"]
-        except requests.exceptions.RequestException as e:
+        except requests.exceptions.HTTPError as e:
             self.render_import_error(e)
+            raise e
 
     def wait_for_execution_results_import_to_finish(self, job_id: str) -> bool:
         try:
             while True:
-                import_status = self.get_job_result(
-                    "executionResultsImporterJob/", job_id
-                )
+                import_status = self.get_job_result("executionResultsImporterJob/", job_id)
                 if import_status is not None:
                     break
                 spin_spinner("Waiting until import of execution results is done")
@@ -280,6 +265,7 @@ class Connection:
                 raise AssertionError(result)
         except requests.exceptions.RequestException as e:
             self.render_import_error(e)
+            raise e
 
     def render_import_error(self, e):
         pprint("!!!ERROR DURING IMPORT!!!", style="#ff0e0e italic")
@@ -289,9 +275,7 @@ class Connection:
         pprint(f"URL: {e.response.url}")
 
     def get_imp_job_result(self, job_id):
-        report_import_status = self.get_job_result(
-            "executionResultsImporterJob/", job_id
-        )
+        report_import_status = self.get_job_result("executionResultsImporterJob/", job_id)
         if report_import_status is None:
             return None
         result = report_import_status["result"]
@@ -302,16 +286,56 @@ class Connection:
 
     def get_test_cycle_structure(self, cycle_key: str) -> List[dict]:
         test_cycle_structure = self.session.get(
-            self.server_url + "cycle/" + cycle_key + "/structure",
+            f"{self.server_url}cycle/{cycle_key}/structure",
         )
         return test_cycle_structure.json()
 
     def get_tov_structure(self, tovKey: str) -> List[dict]:
         tov_structure = self.session.get(
-            self.server_url + "tov/" + tovKey + "/structure",
+            f"{self.server_url}tov/{tovKey}/structure",
         )
         return tov_structure.json()
 
+    def get_test_cases(self, test_case_set_structure: Dict[str, Any]) -> Dict[str, Dict]:
+        spec_test_cases = self.get_spec_test_cases(
+            test_case_set_structure["TestCaseSet_structure"]["key"]["serial"],
+            test_case_set_structure["spec"]["Specification_key"]["serial"],
+        )
+        test_cases = {tc["uniqueID"]: tc for tc in spec_test_cases}
+        if not test_case_set_structure.get("exec"):
+            return {"spec": test_cases}
+        exec_test_cases = self.get_exec_test_cases(
+            test_case_set_structure["TestCaseSet_structure"]["key"]["serial"],
+            test_case_set_structure["exec"]["Execution_key"]["serial"],
+        )
+        test_cases_execs = {tc["uniqueID"]: tc for tc in exec_test_cases}
+        equal_lists = False not in [
+            test_cases.get(uid, {}).get('testCaseSpecificationKey')['serial']
+            == tc['paramCombPK']['serial']
+            for uid, tc in test_cases_execs.items()
+        ]
+        return {
+            "spec": test_cases,
+            "exec": test_cases_execs,
+            "equal_lists": equal_lists,
+        }
+
+    def get_spec_test_cases(self, testCaseSetKey: str, specificationKey: str) -> List[dict]:
+        spec_test_cases = self.session.get(
+            f"{self.server_url}testCaseSets/"
+            f"{testCaseSetKey}/specifications/"
+            f"{specificationKey}/testCases",
+        )
+        return spec_test_cases.json()
+
+    def get_exec_test_cases(self, testCaseSetKey: str, executionKey: str) -> List[dict]:
+        exec_test_cases = self.session.get(
+            f"{self.server_url}testCaseSets/"
+            f"{testCaseSetKey}/executions/"
+            f"{executionKey}/testCases",
+        )
+        return exec_test_cases.json()
+
 
 def login(server="", login="", pwd="") -> Connection:
     if server and login and pwd:
@@ -331,7 +355,8 @@ def login(server="", login="", pwd="") -> Connection:
                 return connection
 
         except requests.HTTPError:
-            print("Invalid login credentials.")
+            logger.error("Invalid login credentials.")
+            logger.debug(traceback.format_exc())
             action = questions.ask_for_action_after_failed_login()
             if action == "retry_password":
                 credentials["password"] = questions.ask_for_testbench_password()
@@ -344,7 +369,8 @@ def login(server="", login="", pwd="") -> Connection:
                 close_program()
 
         except (requests.ConnectionError, requests.exceptions.MissingSchema):
-            print("Invalid server url.")
+            logger.error("Invalid server url.")
+            logger.debug(traceback.format_exc())
             action = questions.ask_for_action_after_failed_server_connection()
             if action == "retry_server":
                 credentials["server_url"] = questions.ask_for_test_bench_server_url()
@@ -354,7 +380,8 @@ def login(server="", login="", pwd="") -> Connection:
                 close_program()
 
         except requests.exceptions.Timeout:
-            print("No connection could be established due to timeout.")
+            logger.error("No connection could be established due to timeout.")
+            logger.debug(traceback.format_exc())
             action = questions.ask_for_action_after_login_timeout()
             if action == "retry":
                 pass
@@ -402,10 +429,10 @@ class ConnectionLog:
         self.connections.append(new_connection)
 
     def export_as_json(self, output_file_path: str):
-        print("Generating JSON export")
-        export_dict = {
-            "configuration": [connection.export() for connection in self.connections]
-        }
+        logger.info("Generating JSON export")
+        export_config = CliReporterConfig(
+            configuration=[connection.export() for connection in self.connections]
+        )
 
         with open(output_file_path, "w") as output_file:
-            json.dump(export_dict, output_file, indent=2)
+            json.dump(dataclasses.asdict(export_config), output_file, indent=2)
